@@ -15,7 +15,7 @@ use tokio::{
     sync::{mpsc, watch},
     time::timeout,
 };
-use tracing::{trace, warn};
+use tracing::{error, trace, warn};
 
 // TODO: this seems arbitrary - it may need tuned or made configurable.
 const INCOMING_MESSAGE_BUFFER_LEN: usize = 10_000;
@@ -23,6 +23,9 @@ const INCOMING_MESSAGE_BUFFER_LEN: usize = 10_000;
 // TODO: this seems arbitrary - it may need tuned or made configurable.
 const ENDPOINT_VERIFICATION_TIMEOUT: Duration = Duration::from_secs(30);
 
+// Error reason for closing a connection when triggered manually by qp2p apis
+const QP2P_CLOSED_CONNECTION: &str = "The connection was closed intentionally by qp2p.";
+
 /// The sending API for a connection.
 #[derive(Clone)]
 pub struct Connection {
@@ -108,7 +111,10 @@ impl Connection {
                             .await
                             .map_err(|error| match &error {
                                 // don't retry on connection loss, since we can't recover that from here
-                                SendError::ConnectionLost(_) => backoff::Error::Permanent(error),
+                                SendError::ConnectionLost(_) => {
+                                    error!("Connection failed on send {:?}", error);
+                                    backoff::Error::Permanent(error)
+                                }
                                 _ => backoff::Error::Transient(error),
                             })
                     })
@@ -146,11 +152,11 @@ impl Connection {
     /// [`ConnectionError::Closed`]`(`[`Close::Local`]`)`, and data on unfinished streams is not
     /// guaranteed to be delivered.
     pub fn close(&self, reason: Option<String>) {
-        let reason = reason
-            .unwrap_or_else(|| "The connection was closed intentionally by qp2p.".to_string());
+        let reason = reason.unwrap_or_else(|| QP2P_CLOSED_CONNECTION.to_string());
         self.inner.close(0u8.into(), &reason.into_bytes());
     }
 
+    /// Opens a uni directional stream and sends message on this stream
     async fn send_uni(&self, msg: Bytes, priority: i32) -> Result<(), SendError> {
         let mut send_stream = self.open_uni().await.map_err(SendError::ConnectionLost)?;
         send_stream.set_priority(priority);
@@ -531,7 +537,7 @@ async fn handle_endpoint_verification(
             addr
         );
         let connection = endpoint
-            .connect(&addr, SERVER_NAME)
+            .connect(addr, SERVER_NAME)
             .map_err(ConnectionError::from)?
             .await?;
 
@@ -591,6 +597,7 @@ where
             Err(error) => {
                 let error = error.into();
                 if error.is_benign() {
+                    warn!("Benign error ignored {:?}", error);
                     None
                 } else {
                     Some(Err(error))
@@ -612,6 +619,7 @@ mod tests {
     use bytes::Bytes;
     use color_eyre::eyre::{bail, Result};
     use futures::{StreamExt, TryStreamExt};
+    use quinn::Endpoint as QuinnEndpoint;
     use std::time::Duration;
 
     #[tokio::test]
@@ -619,23 +627,17 @@ mod tests {
     async fn basic_usage() -> Result<()> {
         let config = InternalConfig::try_from_config(Default::default())?;
 
-        let mut builder = quinn::Endpoint::builder();
-        let _ = builder
-            .listen(config.server.clone())
-            .default_client_config(config.client.clone());
-        let (peer1, _) = builder.bind(&local_addr())?;
+        let (mut peer1, _peer1_incoming) =
+            QuinnEndpoint::server(config.server.clone(), local_addr())?;
+        peer1.set_default_client_config(config.client);
 
-        let mut builder = quinn::Endpoint::builder();
-        let _ = builder
-            .listen(config.server.clone())
-            .default_client_config(config.client.clone());
-        let (peer2, peer2_incoming) = builder.bind(&local_addr())?;
+        let (peer2, peer2_incoming) = QuinnEndpoint::server(config.server.clone(), local_addr())?;
 
         {
             let (p1_tx, mut p1_rx) = Connection::new(
                 peer1.clone(),
                 None,
-                peer1.connect(&peer2.local_addr()?, SERVER_NAME)?.await?,
+                peer1.connect(peer2.local_addr()?, SERVER_NAME)?.await?,
             );
 
             let (p2_tx, mut p2_rx) =
@@ -685,23 +687,17 @@ mod tests {
             ..Default::default()
         })?;
 
-        let mut builder = quinn::Endpoint::builder();
-        let _ = builder
-            .listen(config.server.clone())
-            .default_client_config(config.client.clone());
-        let (peer1, _) = builder.bind(&local_addr())?;
+        let (mut peer1, _peer1_incoming) =
+            QuinnEndpoint::server(config.server.clone(), local_addr())?;
+        peer1.set_default_client_config(config.client);
 
-        let mut builder = quinn::Endpoint::builder();
-        let _ = builder
-            .listen(config.server.clone())
-            .default_client_config(config.client.clone());
-        let (peer2, peer2_incoming) = builder.bind(&local_addr())?;
+        let (peer2, peer2_incoming) = QuinnEndpoint::server(config.server.clone(), local_addr())?;
 
         // open a connection between the two peers
         let (p1_tx, _) = Connection::new(
             peer1.clone(),
             None,
-            peer1.connect(&peer2.local_addr()?, SERVER_NAME)?.await?,
+            peer1.connect(peer2.local_addr()?, SERVER_NAME)?.await?,
         );
 
         let (_, mut p2_rx) =
@@ -730,26 +726,20 @@ mod tests {
     }
 
     #[tokio::test]
-    async fn endpoint_echo() -> Result<()> {
-        let config = InternalConfig::try_from_config(Default::default())?;
+    async fn test_endpoint_echo() -> Result<()> {
+        let config = InternalConfig::try_from_config(Config::default())?;
 
-        let mut builder = quinn::Endpoint::builder();
-        let _ = builder
-            .listen(config.server.clone())
-            .default_client_config(config.client.clone());
-        let (peer1, _) = builder.bind(&local_addr())?;
+        let (mut peer1, _peer1_incoming) =
+            QuinnEndpoint::server(config.server.clone(), local_addr())?;
+        peer1.set_default_client_config(config.client);
 
-        let mut builder = quinn::Endpoint::builder();
-        let _ = builder
-            .listen(config.server.clone())
-            .default_client_config(config.client.clone());
-        let (peer2, peer2_incoming) = builder.bind(&local_addr())?;
+        let (peer2, peer2_incoming) = QuinnEndpoint::server(config.server.clone(), local_addr())?;
 
         {
             let (p1_tx, _) = Connection::new(
                 peer1.clone(),
                 None,
-                peer1.connect(&peer2.local_addr()?, SERVER_NAME)?.await?,
+                peer1.connect(peer2.local_addr()?, SERVER_NAME)?.await?,
             );
 
             // we need to accept the connection on p2, or the message won't be processed
@@ -789,23 +779,19 @@ mod tests {
     async fn endpoint_verification() -> Result<()> {
         let config = InternalConfig::try_from_config(Default::default())?;
 
-        let mut builder = quinn::Endpoint::builder();
-        let _ = builder
-            .listen(config.server.clone())
-            .default_client_config(config.client.clone());
-        let (peer1, peer1_incoming) = builder.bind(&local_addr())?;
+        let (mut peer1, peer1_incoming) =
+            QuinnEndpoint::server(config.server.clone(), local_addr())?;
+        peer1.set_default_client_config(config.client.clone());
 
-        let mut builder = quinn::Endpoint::builder();
-        let _ = builder
-            .listen(config.server.clone())
-            .default_client_config(config.client.clone());
-        let (peer2, peer2_incoming) = builder.bind(&local_addr())?;
+        let (mut peer2, peer2_incoming) =
+            QuinnEndpoint::server(config.server.clone(), local_addr())?;
+        peer2.set_default_client_config(config.client);
 
         {
             let (p1_tx, _) = Connection::new(
                 peer1.clone(),
                 None,
-                peer1.connect(&peer2.local_addr()?, SERVER_NAME)?.await?,
+                peer1.connect(peer2.local_addr()?, SERVER_NAME)?.await?,
             );
 
             // we need to accept the connection on p2, or the message won't be processed
