@@ -10,19 +10,11 @@
 use super::{hash, local_addr, new_endpoint, random_msg};
 use crate::{Config, Endpoint, RetryConfig};
 use color_eyre::eyre::{bail, eyre, Report, Result};
-use futures::{
-    future,
-    stream::{self, FuturesUnordered},
-    StreamExt, TryStreamExt,
-};
-use std::{
-    collections::BTreeSet,
-    sync::atomic::{AtomicUsize, Ordering},
-    time::Duration,
-};
-use tokio::sync::Notify;
+use futures::future;
+use std::sync::Arc;
+use std::{collections::BTreeSet, time::Duration};
 use tracing::info;
-use tracing_test::traced_test;
+// use tracing_test::traced_test;
 
 #[tokio::test(flavor = "multi_thread")]
 async fn successful_connection() -> Result<()> {
@@ -217,7 +209,7 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
             assert_eq!(connection.remote_address(), alice_addr);
             incoming
         } else {
-            bail!("No incoming connectino from Alice");
+            bail!("No incoming connection from Alice");
         };
 
     let msg0 = random_msg(1024);
@@ -272,55 +264,64 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
     let (bob, mut bob_incoming_connections, _) = new_endpoint().await?;
     let bob_addr = bob.public_addr();
 
-    // Try to establish two connections to the same peer at the same time.
-    let ((b_to_a, mut bob_incoming_messages), (_, _)) =
-        future::try_join(bob.connect_to(&alice_addr), bob.connect_to(&alice_addr)).await?;
+    let (carol, mut carol_incoming_connections, _) = new_endpoint().await?;
+    let carol_addr = carol.public_addr();
 
-    // Alice gets both incoming connections
-    let (a_to_b, mut alice_incoming_messages) = if let Ok(Some((connection, incoming))) =
-        alice_incoming_connections.next().timeout().await
-    {
-        assert_eq!(connection.remote_address(), bob_addr);
-        (connection, incoming)
-    } else {
-        bail!("Missing incoming connection");
-    };
-    if let Ok(Some((connection, _))) = alice_incoming_connections.next().timeout().await {
-        assert_eq!(connection.remote_address(), bob_addr);
-    } else {
-        bail!("Missing incoming connection");
-    };
-
-    if let Ok(Some((connection, _))) = alice_incoming_connections.next().timeout().await {
-        bail!(
-            "Unexpected incoming connection from {}",
-            connection.remote_address()
-        );
-    }
-
-    // Send two messages, one from each peer
     let msg0 = random_msg(1024);
-    a_to_b.send(msg0.clone()).await?;
-
     let msg1 = random_msg(1024);
-    b_to_a.send(msg1.clone()).await?;
+    let msg2 = random_msg(1024);
 
-    // Bob did not get a new incoming connection
-    if let Ok(Some(_)) = bob_incoming_connections.next().timeout().await {
-        bail!("Unexpected incoming connection from alice to bob");
-    }
+    let (b_to_a, mut b_incoming_from_a) = bob.connect_to(&alice_addr).await?;
+    let (c_to_a, mut c_incoming_from_a) = carol.connect_to(&alice_addr).await?;
 
-    // Both messages are received  at the other end
-    if let Ok(message) = alice_incoming_messages.next().timeout().await {
-        assert_eq!(message?, Some(msg1));
-    } else {
-        bail!("No message received from Bob");
-    }
+    if let Some((a_to_b, mut alice_incoming_messages)) = alice_incoming_connections.next().await {
+        if let Some((a_to_c, mut alice_incoming_messages_2)) =
+            alice_incoming_connections.next().await
+        {
+            // all cehcks need to happen within here as conns are kept alive inside this if block
 
-    if let Ok(message) = bob_incoming_messages.next().timeout().await {
-        assert_eq!(message?, Some(msg0));
-    } else {
-        bail!("No message from alice");
+            assert_eq!(a_to_b.remote_address(), bob_addr);
+            assert_eq!(a_to_c.remote_address(), carol_addr);
+
+            b_to_a.send(msg0.clone()).await?;
+            c_to_a.send(msg1.clone()).await?;
+
+            // check bob hasnt been connected to for some reason
+            if let Ok(Some(_)) = bob_incoming_connections.next().timeout().await {
+                bail!("Unexpected incoming connection from alice to bob");
+            }
+            // check carol hasnt been connected to for some reason
+            if let Ok(Some(_)) = carol_incoming_connections.next().timeout().await {
+                bail!("Unexpected incoming connection from alice to carol");
+            }
+
+            // Both messages are received  at the alice end
+            if let Ok(message) = alice_incoming_messages.next().timeout().await {
+                assert_eq!(message?, Some(msg0));
+            } else {
+                bail!("No message received from Bob");
+            }
+
+            if let Ok(message) = alice_incoming_messages_2.next().timeout().await {
+                assert_eq!(message?, Some(msg1));
+            } else {
+                bail!("No message from carol");
+            }
+
+            a_to_b.send(msg2.clone()).await?;
+            a_to_c.send(msg2.clone()).await?;
+
+            if let Ok(message) = b_incoming_from_a.next().timeout().await {
+                assert_eq!(message?, Some(msg2.clone()));
+            } else {
+                bail!("No message received from Alice to bob");
+            }
+            if let Ok(message) = c_incoming_from_a.next().timeout().await {
+                assert_eq!(message?, Some(msg2));
+            } else {
+                bail!("No message received from Alive to Carol");
+            }
+        }
     }
 
     Ok(())
@@ -330,15 +331,20 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
 async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
     use futures::future;
 
-    let num_senders: usize = 10;
+    let num_senders: usize = 100;
     let num_messages_each: usize = 100;
-    let num_messages_total: usize = 1000;
+    let num_messages_total: usize = num_senders * num_messages_each;
 
     let (server_endpoint, mut recv_incoming_connections, _) = new_endpoint().await?;
     let server_addr = server_endpoint.public_addr();
+    let msg_len = 200;
+    let test_msgs: Vec<_> = (0..num_messages_each)
+        .map(|_| random_msg(msg_len))
+        .collect();
+    assert_eq!(test_msgs.len(), num_messages_each);
+    assert_eq!(msg_len, test_msgs[0].len());
 
-    let test_msgs: Vec<_> = (0..num_messages_each).map(|_| random_msg(1024)).collect();
-    let sending_msgs = test_msgs.clone();
+    let sending_msgs = Arc::new(test_msgs);
 
     let mut tasks = Vec::new();
 
@@ -347,7 +353,7 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
         async move {
             let mut num_received = 0;
             let mut sending_tasks = Vec::new();
-
+            let mut task_connection: Option<super::Connection> = None;
             while let Some((connection, mut recv_incoming_messages)) =
                 recv_incoming_connections.next().await
             {
@@ -357,17 +363,18 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
                         connection.remote_address(),
                         msg.len()
                     );
-                    assert_eq!(msg.len(), test_msgs[0].len());
-
+                    task_connection = Some(connection.clone());
                     sending_tasks.push(tokio::spawn({
                         let connection = connection.clone();
                         async move {
-                            // Hash the inputs for couple times to simulate certain workload.
                             let hash_result = hash(&msg);
+                            // to simulate certain workload.
                             for _ in 0..5 {
                                 let _ = hash(&msg);
                             }
+
                             // Send the hash result back.
+                            info!("About to send hash from receiver");
                             connection.send(hash_result.to_vec().into()).await?;
 
                             Ok::<_, Report>(())
@@ -381,12 +388,11 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
                 }
             }
 
-            future::try_join_all(sending_tasks)
-                .await?
-                .into_iter()
-                .collect::<Result<_>>()?;
+            let _res = future::try_join_all(sending_tasks).await?;
 
-            Ok(())
+            info!("Receiver closed");
+
+            Ok(task_connection)
         }
     }));
 
@@ -395,12 +401,15 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
         let messages = sending_msgs.clone();
         tasks.push(tokio::spawn({
             let (send_endpoint, _, _) = new_endpoint().await?;
+            let task_connection: Option<super::Connection> = None;
 
             async move {
                 let mut hash_results = BTreeSet::new();
                 let (connection, mut recv_incoming_messages) =
                     send_endpoint.connect_to(&server_addr).await?;
-                for (index, message) in messages.iter().enumerate().take(num_messages_each) {
+
+                assert_eq!(messages.len(), num_messages_each);
+                for (index, message) in messages.iter().enumerate() {
                     let _ = hash_results.insert(hash(message));
                     info!("sender #{} sending message #{}", id, index);
                     connection.send(message.clone()).await?;
@@ -410,6 +419,7 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
                     "sender #{} completed sending messages, starts listening",
                     id
                 );
+                let mut all_received_msgs = 0;
 
                 while let Ok(Ok(Some(msg))) = recv_incoming_messages.next().timeout().await {
                     info!(
@@ -422,133 +432,106 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
                     info!("Hash len before: {:?}", hash_results.len());
                     assert!(hash_results.remove(&msg[..]));
                     info!("Hash len after: {:?}", hash_results.len());
-
-                    if hash_results.is_empty() {
+                    all_received_msgs += 1;
+                    if hash_results.is_empty() && all_received_msgs == num_messages_each {
                         break;
                     }
                 }
 
-                Ok::<_, Report>(())
+                Ok::<_, Report>(task_connection)
             }
         }));
     }
 
-    future::try_join_all(tasks)
-        .await?
-        .into_iter()
-        .collect::<Result<_>>()?;
+    let _res = future::try_join_all(tasks).await?;
 
     Ok(())
 }
 
 #[tokio::test(flavor = "multi_thread")]
-#[traced_test]
 async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<()> {
+    use futures::future;
+
     let num_senders: usize = 10;
     let num_messages_each: usize = 100;
     let num_messages_total: usize = num_senders * num_messages_each;
 
-    let (server_endpoint, recv_incoming_connections, _) = new_endpoint().await?;
+    let (server_endpoint, mut recv_incoming_connections, _) = new_endpoint().await?;
     let server_addr = server_endpoint.public_addr();
-
     let msg_len = 1024 * 1024;
     let test_msgs: Vec<_> = (0..num_messages_each)
         .map(|_| random_msg(msg_len))
         .collect();
-    let sending_msgs = test_msgs.clone();
+    assert_eq!(test_msgs.len(), num_messages_each);
+    assert_eq!(msg_len, test_msgs[0].len());
 
-    let mut tasks = FuturesUnordered::new();
+    let sending_msgs = Arc::new(test_msgs);
 
-    // Receiver + Hasher
-    tasks.push(tokio::spawn(async move {
-        let num_received = AtomicUsize::new(0);
-        assert!(!logs_contain("error"));
-
-        let finished = Notify::new();
-
-        let recv_incoming_connections = stream::unfold(
-            recv_incoming_connections,
-            |mut recv_incoming_connections| {
-                let finished = &finished;
-                async move {
-                    tokio::select! {
-                        Some(connection) = recv_incoming_connections.next() => Some((connection, recv_incoming_connections)),
-                        _ = finished.notified() => {
-                            info!("finished");
-                            None
-                        },
-                        else => None,
-                    }
-                }
-            },
-        );
-
-        recv_incoming_connections
-            .map(Result::<_, Report>::Ok)
-            .try_for_each_concurrent(
-                None,
-                |(connection, mut recv_incoming_messages)| {
-                    let num_received = &num_received;
-                    let finished = &finished;
-                    async move {
-                        let mut last_count = 0;
-
-                        while let Some(msg) = recv_incoming_messages.next().timeout().await?? {
-                            info!(
-                                "received from {:?} with message size {}",
-                                connection.remote_address(),
-                                msg.len()
-                            );
-                            assert!(!logs_contain("error"));
-
-                            assert_eq!(msg.len(), msg_len);
+    let mut tasks = Vec::new();
 
+    // Receiver
+    tasks.push(tokio::spawn({
+        async move {
+            let mut num_received = 0;
+            let mut sending_tasks = Vec::new();
+            let mut task_connection: Option<super::Connection> = None;
+            while let Some((connection, mut recv_incoming_messages)) =
+                recv_incoming_connections.next().await
+            {
+                while let Ok(Ok(Some(msg))) = recv_incoming_messages.next().timeout().await {
+                    info!(
+                        "received from {:?} with message size {}",
+                        connection.remote_address(),
+                        msg.len()
+                    );
+                    task_connection = Some(connection.clone());
+                    sending_tasks.push(tokio::spawn({
+                        let connection = connection.clone();
+                        async move {
                             let hash_result = hash(&msg);
+                            // to simulate certain workload.
                             for _ in 0..5 {
                                 let _ = hash(&msg);
                             }
 
                             // Send the hash result back.
+                            info!("About to send hash from receiver");
                             connection.send(hash_result.to_vec().into()).await?;
 
-                            assert!(!logs_contain("error"));
-
-                            last_count = num_received.fetch_add(1, Ordering::SeqCst) + 1;
-                            info!("counted {}/{} messages so far", last_count, num_messages_total);
+                            Ok::<_, Report>(())
                         }
+                    }));
 
-                        if last_count >= num_messages_total {
-                            info!("counted all {} messages, finishing", num_messages_total);
-                            finished.notify_one();
-                        }
+                    num_received += 1;
+                }
+                if num_received >= num_messages_total {
+                    break;
+                }
+            }
 
-                        Ok(())
-                    }
-                },
-            )
-            .await?;
+            let _res = future::try_join_all(sending_tasks).await?;
 
-        assert!(!logs_contain("error"));
+            info!("RECEIVER CLOSED");
 
-        Ok(())
+            Ok(task_connection)
+        }
     }));
 
-    // Sender + Verifier
+    // Sender
     for id in 0..num_senders {
         let messages = sending_msgs.clone();
-        assert!(!logs_contain("error"));
-
         tasks.push(tokio::spawn({
             let (send_endpoint, _, _) = new_endpoint().await?;
+            let task_connection: Option<super::Connection> = None;
 
             async move {
                 let mut hash_results = BTreeSet::new();
-
                 let (connection, mut recv_incoming_messages) =
                     send_endpoint.connect_to(&server_addr).await?;
-                for (index, message) in messages.iter().enumerate().take(num_messages_each) {
-                    let _ = hash_results.insert(hash(message));
 
+                assert_eq!(messages.len(), num_messages_each);
+                for (index, message) in messages.iter().enumerate() {
+                    let _ = hash_results.insert(hash(message));
                     info!("sender #{} sending message #{}", id, index);
                     connection.send(message.clone()).await?;
                 }
@@ -558,11 +541,10 @@ async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<(
                     id
                 );
 
-                while let Some(msg) = recv_incoming_messages.next().await? {
-                    assert!(!logs_contain("error"));
-
+                let mut all_received_msgs = 0;
+                while let Ok(Ok(Some(msg))) = recv_incoming_messages.next().timeout().await {
                     info!(
-                        "sender #{} received from server {:?} with message size {:?}",
+                        "#{} received from server {:?} with message size {}",
                         id,
                         connection.remote_address(),
                         msg.len()
@@ -572,27 +554,18 @@ async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<(
                     assert!(hash_results.remove(&msg[..]));
                     info!("Hash len after: {:?}", hash_results.len());
 
-                    if hash_results.is_empty() {
+                    all_received_msgs += 1;
+                    if hash_results.is_empty() && all_received_msgs == num_messages_each {
                         break;
                     }
                 }
 
-                assert!(
-                    hash_results.is_empty(),
-                    "verifier terminated before all results were verified"
-                );
-
-                Ok::<_, Report>(())
+                Ok::<_, Report>(task_connection)
             }
         }));
     }
 
-    while let Some(result) = tasks.next().await {
-        match result {
-            Ok(Ok(())) => (),
-            other => bail!("Error from test threads: {:?}", other),
-        }
-    }
+    let _res = future::try_join_all(tasks).await?;
 
     Ok(())
 }
@@ -665,7 +638,6 @@ async fn many_messages() -> Result<()> {
 // that succeeds. We should still be able to establish a connection with the rest of the
 // bootstrap contacts later.
 #[tokio::test(flavor = "multi_thread")]
-#[traced_test]
 async fn connection_attempts_to_bootstrap_contacts_should_succeed() -> Result<()> {
     let (ep1, _ep1_incoming, _) = new_endpoint().await?;
     let (ep2, _ep2_incoming, _) = new_endpoint().await?;
@@ -673,7 +645,7 @@ async fn connection_attempts_to_bootstrap_contacts_should_succeed() -> Result<()
 
     let contacts = vec![ep1.public_addr(), ep2.public_addr(), ep3.public_addr()];
 
-    let (ep, _, bootstrapped_peer) = Endpoint::new(
+    let (ep, _, bootstrapped_peer) = Endpoint::new_peer(
         local_addr(),
         &contacts,
         Config {
@@ -712,7 +684,6 @@ async fn reachability() -> Result<()> {
 }
 
 #[tokio::test(flavor = "multi_thread")]
-#[traced_test]
 async fn client() -> Result<()> {
     use crate::{Config, Endpoint};
 
@@ -735,6 +706,7 @@ async fn client() -> Result<()> {
         .next()
         .await
         .ok_or_else(|| eyre!("did not receive expected connection"))?;
+
     assert_eq!(server_to_client.remote_address(), client.public_addr());
 
     let message = server_messages
@@ -762,6 +734,9 @@ async fn client() -> Result<()> {
     drop(server_to_client);
     drop(server_messages);
 
+    // small way for all drop effects to propagate
+    tokio::time::sleep(tokio::time::Duration::from_millis(100)).await;
+
     // sending should now fail, since the connection was closed at the peer
     match client_to_server.send(b"world"[..].into()).await {
         Err(crate::SendError::ConnectionLost(_)) => {}
